#include "Thread.hh"
#include "ThreadError.hh"
#include "Semaphore.hh"
#include "Mutex.hh"
#include "Guard.hh"

#include "WorkItem.hh"
#include "Queue.hh"

#include <iostream>
#include <ctime>
#include <vector>

using namespace std;

class Base: public Thread {
	bool stopped_;
	Mutex m_;

	const int id_;
	Queue& q_;
public:
	Base(int id, Queue& q)
	: Thread(true),
	  stopped_(false),
	  id_(id),
	  q_(q)
	{}
	void stop() {
		Guard guard(m_);
		stopped_=true;
	}
	bool is_stopped() {
		Guard guard(m_);
		return stopped_;
	}
	int get_id() const {
		return id_;
	}
	Queue& get_queue() const {
		return q_;
	}
};

class P:public Base {
public:
	P(int id, Queue& q)
	: Base(id,q)
	{}

	void run() {
		while(1) {
			if(is_stopped()){
				cerr << "P(" << get_id() 
					<< ") stopped..." << endl;
				return;
			}

			WorkItem work=WorkItem::produce();

			cerr << "P(" << get_id() 
				<< ") going to put... queue size is " 
				<< get_queue().size() << endl;
			get_queue().put(work);
			cerr << "P(" << get_id() << ") put..." << endl;
		}
	}
};

class C:public Base {
public:
	C(int id, Queue& q)
	: Base(id,q)
	{}

	void run() {
		while(1) {
			if(is_stopped()) {
				cerr << "C(" << get_id() 
					<< ") stopped..." << endl;
				return;
			}

			cerr << "C(" << get_id() 
				<< ") going to get... queue size is " 
				<< get_queue().size() << endl;
			WorkItem work=get_queue().get();
			cerr << "C(" << get_id() << ") get..." << endl;
			work.consume();
		}
	}
};

static const int PRODUCERS_COUNT=1;
static const int CONSUMERS_COUNT=1;

int main() {
	Queue q(3);

	vector<P*> producers;
	vector<C*> consumers;
	
	//creating threads
	for(int i=0;i<PRODUCERS_COUNT;i++) {
		producers.push_back(new P(i,q));
	}
	for(int i=0;i<CONSUMERS_COUNT;i++) {
		consumers.push_back(new C(i,q));
	}
	
	//starting threads
	for(vector<C*>::iterator it=consumers.begin();
				it!=consumers.end();++it){
		(*it)->start();
	}
	for(vector<P*>::iterator it=producers.begin();
				it!=producers.end();++it){
		(*it)->start();
	}

	sleep(20);

	// stopping threads
	for(vector<P*>::iterator it=producers.begin();
				it!=producers.end();++it){
		(*it)->stop();
	}
	for(vector<C*>::iterator it=consumers.begin();
				it!=consumers.end();++it){
		(*it)->stop();
	}


	// signaling consumers
	while(q.is_empty()) {
		cerr << "signaling consumers..." << endl;
		q.put(WorkItem::produce());
		sleep(1);
	}
	// signaling producers
	while(q.is_full()) {
		cerr << "signaling producers..." << endl;
		q.get();
		sleep(1);
	}

	// joining producers	
	for(vector<P*>::iterator it=producers.begin();
				it!=producers.end();++it){
		cerr << "joining producers..." << endl;
		(*it)->join();
	}
	
	
	// joining consumers
	for(vector<C*>::iterator it=consumers.begin();
				it!=consumers.end();++it){
		cerr << "joining consumers..." << endl;
		(*it)->join();
	}
	
	// deleting threads
	for(vector<P*>::iterator it=producers.begin();
				it!=producers.end();++it){
		cerr << "deleting producers..." << endl;
		delete (*it);
	}
	for(vector<C*>::iterator it=consumers.begin();
				it!=consumers.end();++it){
		cerr << "deleting consumers..." << endl;
		delete (*it);
	}

	return 0;
};



